/**
    coordinates.h
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#ifndef __COORDINATES_H__
#define __COORDINATES_H__

namespace mkr { namespace formation {

/**
 * @brief Class Coordinates provides an object which collects all coordinates of the robot
 */
class Coordinates {
public:

    /**
     * @brief constructor which sets all initial coordinates to 0
     */
    Coordinates(void);

    /**
     * @brief constructor
     *
     * @param x - x coordinate
     * @param y - y coordinate
     * @param z - z coordinate
     * @param phi - heading connected with the kinematic model of the leader
     * @param psi - angle between the camera optical axis and the x-axis in xy-plane
     * @param eta - angle between the camera optical axis and the xy-plane
     */
    Coordinates(const double x, const double y, const double z, const double phi, const double psi, const double eta);

    /**
     * @brief destructor
     */
    ~Coordinates(void);

    double x;
    double y;
    double z;
    double phi;
    double psi;
    double eta;
};

}}

#endif

