/**
    angularsolver.cc

    @author Vit Kratky
    @version 1.3 20/05/17
*/

#include <iostream>
#include "angularsolver.h"
extern "C" {
int x_is_newA;
double objrepA,objepsA,gLgepsA;
#include "cfsqpusr.h"
}

using namespace std;

AngularSolver::AngularSolver(const double *actualState, double *x, const double angMax, const double pitchMax,
                             vCoordinates poi, int counter)
    : actualState(actualState), x(x), angMax(angMax), pitchMax(pitchMax), poi(poi), counter(counter){
    states = (double *)calloc((N_PARAMS_F+1)*LENGTH_HORIZONT,sizeof(double));//+curvature
    callSolver();
}

AngularSolver::~AngularSolver() {
    free(states);
}

void AngularSolver::generateBoundaryValues(double * bl,double * bu){
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        bl[j] = -angMax;
        bl[LENGTH_HORIZONT+j] = -pitchMax;
        bu[j] = angMax;
        bu[LENGTH_HORIZONT+j] = pitchMax;
    }
}

void  AngularSolver::setState(double *x, double stateX, double stateY, double stateZ, double statePhi, int j){
        if (fabs(x[LENGTH_HORIZONT+j]) < (1E-6F)) { //if K==0
            states[j]= stateX + x[j]*cos(statePhi)*TS;
            states[LENGTH_HORIZONT+j]= stateY + x[j]*sin(statePhi)*TS;
        } else {
            states[j]= stateX + (-sin(statePhi)+sin(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
            states[LENGTH_HORIZONT+j]= stateY - (-cos(statePhi)+cos(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
        }
        states[2*LENGTH_HORIZONT+j]= statePhi + x[LENGTH_HORIZONT+j]*x[j]*TS;
        states[4*LENGTH_HORIZONT+j] = stateZ + x[2*LENGTH_HORIZONT+j]*TS;
}

void AngularSolver::computeStatesPosition(double *x){
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        if (j == 0) setState(x, actualState[0], actualState[1], actualState[7], actualState[2], j);
        else setState(x, states[j-1], states[LENGTH_HORIZONT+j-1], states[4*LENGTH_HORIZONT+j-1],
                states[2*LENGTH_HORIZONT+j-1], j);

    }
}

void AngularSolver::computeStates(double *x){
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        if (j == 0){
            states[3*LENGTH_HORIZONT] = fmod(actualState[3] + x[j]*TS + 2*PI, 2*PI);
            states[5*LENGTH_HORIZONT] = actualState[9] + x[LENGTH_HORIZONT+j]*TS;
        }
        else{
            states[3*LENGTH_HORIZONT+j] = fmod(states[3*LENGTH_HORIZONT+j-1] + x[j]*TS + 2*PI, 2*PI);
            states[5*LENGTH_HORIZONT+j] = states[5*LENGTH_HORIZONT+j-1] + x[LENGTH_HORIZONT+j]*TS;
        }
   }
}

double AngularSolver::getObjectiveFunctionValue(){
    return objf;
}

void objA(int nparam, int j,double *x,double *fj,void *cd) {
    AngularSolver *so = (AngularSolver *) cd;
    so->computeStates(x);
    double phiDiff=0, diffpsi, diffpsi2, omegaDiff=0, epsDiff = 0, chiDiff = 0;
    diffpsi = fabs(so->states[3*LENGTH_HORIZONT] - fmod(atan2(so->poi[so->counter].y -
              so->states[LENGTH_HORIZONT], so->poi[so->counter].x - so->states[0])+2*PI, 2*PI));
    diffpsi2 = diffpsi > PI ? 2*PI-diffpsi : diffpsi;
    phiDiff+=pow(diffpsi2, 2);
    omegaDiff+=pow(so->actualState[6]-x[0], 2);
    epsDiff+=pow(so->actualState[10] - x[LENGTH_HORIZONT], 2);
    chiDiff+=pow(so->states[5*LENGTH_HORIZONT]-atan2(so->poi[so->counter].z - so->states[4*LENGTH_HORIZONT],
            sqrt(pow(so->poi[so->counter].x-so->states[0], 2)+pow(so->poi[so->counter].y-so->states[LENGTH_HORIZONT], 2))), 2);
    for (int i = 1; i < LENGTH_HORIZONT; i++) {
        omegaDiff+=pow(so->states[3*LENGTH_HORIZONT+i]-so->states[3*LENGTH_HORIZONT+i-1], 2);
        diffpsi = fabs(so->states[3*LENGTH_HORIZONT+i] - fmod(atan2(so->poi[so->counter+i].y -
                  so->states[LENGTH_HORIZONT+i], so->poi[so->counter+i].x - so->states[i])+2*PI, 2*PI));
        diffpsi2 = diffpsi > PI ? 2*PI-diffpsi : diffpsi;
        phiDiff+=pow(diffpsi2, 2);
        omegaDiff+=pow(x[i-1]-x[i], 2);
        epsDiff+=pow(x[LENGTH_HORIZONT+i-1] - x[LENGTH_HORIZONT+i], 2);
        chiDiff+=pow(so->states[5*LENGTH_HORIZONT+i]-atan2(so->poi[so->counter+i].z - so->states[4*LENGTH_HORIZONT+i],
                sqrt(pow(so->poi[so->counter+i].x-so->states[i], 2)+pow(so->poi[so->counter+i].y-so->states[LENGTH_HORIZONT+i], 2))), 2);
    }
    *fj=phiDiff+0.05*omegaDiff+0.05*epsDiff+1*chiDiff;
    return;
}

void cntrA(int nparam, int j, double *x, double *gj, void *cd) {
    return;

}

void AngularSolver::callSolver (void) {
    int nparam,nf,nineq,neq,mode,iprint,miter,neqn,nineqn,
            ncsrl,ncsrn,nfsr,mesh_pts[1],inform;
    double bigbnd,eps,epsneq,udelta;
    double *bl,*bu,*f,*g,*lambda,*cd;
    mode=100;
    iprint=0; //printing informations
    miter=500; //maximum number of iterations
    bigbnd=1.e12; // all bounds above bigbnd or under -bigbnd in bu or bl are suposed to be infinite
    eps=1.e-9; //final norm requirement
    epsneq=1.e-6; // maximum allowed violation of nonlinear equality constraints
    udelta=0.e0; //perturbation size
    nparam=2*LENGTH_HORIZONT; //dimension of x
    nf=1; //number of objective functions
    neqn=neq=0; //3; //nonlinear equality constraints / equality constraints
    nineqn=nineq=0; //nonlinear inequality constraints / inequality constraints
    ncsrn=0; //sets of nonlinear sequentially related constraints
    ncsrl=0; //sets of linear sequentially related constraints
    mesh_pts[0]=0;
    nfsr=0; //sequentially related objective functions
    cd=(double *)calloc(nparam,sizeof(double));
    bl=(double *)calloc(nparam,sizeof(double)); //lower bounds
    bu=(double *)calloc(nparam,sizeof(double)); // upper bound
    f=(double *)calloc(nf+1,sizeof(double));
    g=(double *)calloc(nineq+(mesh_pts[0]-1)*(ncsrl+ncsrn)+neq+1,sizeof(double));
    lambda=(double *)calloc(nineq+(mesh_pts[0]-1)*(ncsrl+ncsrn)+neq+nf+nparam,sizeof(double));
    generateBoundaryValues(bl,bu);
    computeStatesPosition(x);
    cfsqp(nparam,nf,nfsr,nineqn,nineq,neqn,neq,ncsrl,ncsrn,mesh_pts,
          mode,iprint,miter,&inform,bigbnd,eps,epsneq,udelta,bl,bu,(x+3*LENGTH_HORIZONT),
          f,g,lambda,objA,cntrA,grobfd,grcnfd,this);
    objA(5,5,(x+3*LENGTH_HORIZONT),&objf,this);
    free(bl);
    free(bu);
    free(f);
    free(g);
    free(lambda);
    free(cd);
}

