/**
    follower.cc
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#include "follower.h"
#include "solver.h"
#include "angularsolver.h"
#include "map.h"

#include <geometry_msgs/Twist.h>
#include <nav_msgs/Odometry.h>
#include <tf/transform_datatypes.h>
#include <geometry_msgs/Pose.h>
#include <math.h>

using namespace mkr::formation;

RobotFollower::RobotFollower(const int type, const double lumInt, const char * name, double * position, const int index,
                             const int nRobots, const double vMax, const double KMax, const double verMax, const double angMax,
                             const double pitchMax, const double lightRatio) : Robot(), lumInt(lumInt), type(type), indexOfRobot(index),
                             nRobots(nRobots), vMax(vMax), verMax(verMax), angMax(angMax), KMax(KMax), pitchMax(pitchMax){
    nIter=N;
    //char numstr[50];
    //string temp = "/home/vitkratky/cvut/project/data/";
    //sprintf(numstr, "follower%d.txt", index);
    //temp=temp+numstr;
    //strcpy(numstr,temp.c_str());
    //statesOutput.open(name);
    //stream.open(numstr);
    x = (double*)calloc(LENGTH_HORIZONT*N_PARAMS_F, sizeof(double));
    followersTrajectories = (double *)calloc((nRobots-2)*(LENGTH_HORIZONT-N)*3, sizeof(double));
    actualState = (double *)calloc(N_STATES, sizeof(double));
    actualPositions = &position[0];
    nh = ros::NodeHandle("~");
    subLeaderActualPosition = nh.subscribe("leader_position", 1, &RobotFollower::leaderPositionCallback, this);
    subThisFollowerActualPosition = nh.subscribe("actual_position", 1, &RobotFollower::thisFollowerPositionCallback, this);
    subFollower2ActualPosition = nh.subscribe("follower2_position", 1, &RobotFollower::secondFollowerPositionCallback, this);
    subFollower3ActualPosition = nh.subscribe("follower3_position", 1, &RobotFollower::thirdFollowerPositionCallback, this);
    subDesiredTrajectory = nh.subscribe("leader_desired_trajectory", 1, &RobotFollower::desiredTrajectoryCallback, this);
    subSyncNumber = nh.subscribe("sync_number", 1, &RobotFollower::syncNumberCallback, this);
    subLeaderPitch = nh.subscribe("leader_pitch", 1, &RobotFollower::leaderPitchCallback, this);
    subFollower2DesiredTrajectory = nh.subscribe("second_follower_desired_trajectory", 1, &RobotFollower::secondFollowerTrajectoryCallback, this);
    subFollower3DesiredTrajectory = nh.subscribe("third_follower_desired_trajectory", 1, &RobotFollower::thirdFollowerTrajectoryCallback, this);
    pubDesiredTrajectory = nh.advertise<mbzirc_trackers::TrackerTrajectory>("desired_trajectory", 1);
    pubDesiredPitch = nh.advertise<mbzirc_trackers::TrackerTrajectory>("desired_pitch", 1);
    pubReady = nh.advertise<std_msgs::Bool>("follower_ready", 1);
    subFormationStart = nh.subscribe("formation_start", 1, &RobotFollower::formationStartCallback, this);
    counter = 0;
    ready = false;
    start = false;
    isTrajectorySet = false;
    input = Controls(0.0, 0.0, 0.0, 0.0, 0.0);
    dist = (-K_L+sqrt(pow(K_L, 2)-4*K_Q*(-lumInt/(DESIRED_ILL*lightRatio) + K_C)))/(2*K_Q);
}

RobotFollower::~RobotFollower() {
    stream.close();
    statesOutput.close();
    free(x);
    free(actualPositions);
    free(actualState);
    free(subscribedTrajectory);
    free(followersTrajectories);
    if(obstacles != NULL){
        free(obstacles);
    }
}

void RobotFollower::desiredTrajectoryCallback(const mbzirc_trackers::TrackerTrajectoryConstPtr &msg){
    mbzirc_trackers::TrackerTrajectory trajectory;
    trajectory.points = msg->points;
    for (int i = 0; i < LENGTH_HORIZONT; i++){
        subscribedTrajectory[i] = trajectory.points[N+i].x;
        subscribedTrajectory[LENGTH_HORIZONT+i] = trajectory.points[N+i].y;
        subscribedTrajectory[2*LENGTH_HORIZONT+i] = fmod(trajectory.points[N+i].yaw+2*PI, 2*PI);
        subscribedTrajectory[3*LENGTH_HORIZONT+i] = trajectory.points[N+i].z;
    }
    bool tmp = isTrajectorySet;
    isTrajectorySet = true;
    if(tmp != isTrajectorySet) {
        cout << "subscribing desired trajectory finished" << endl;
    }
}

void RobotFollower::leaderPitchCallback(const mbzirc_trackers::TrackerTrajectoryConstPtr &msg){
    mbzirc_trackers::TrackerTrajectory trajectory;
    trajectory.points = msg->points;
    for (int i = 0; i < LENGTH_HORIZONT; i++){
        subscribedTrajectory[4*LENGTH_HORIZONT+i] = trajectory.points[N+i].yaw;
    }
}

void RobotFollower::leaderPositionCallback(const nav_msgs::OdometryConstPtr &msg){
    double roll, pitch, yaw;
    actualPositions[0] = msg->pose.pose.position.x;
    actualPositions[1] = msg->pose.pose.position.y;
    actualPositions[7] = msg->pose.pose.position.z;
    actualPositions[3] = fmod(tf::getYaw(msg->pose.pose.orientation)+2*PI, 2*PI);//camera heading
}

void RobotFollower::thisFollowerPositionCallback(const nav_msgs::OdometryConstPtr &msg){
    if(!ready){
    actualState[0] = msg->pose.pose.position.x;
    actualState[1] = msg->pose.pose.position.y;
    actualState[3] = fmod(tf::getYaw(msg->pose.pose.orientation)+2*PI, 2*PI);//camera heading
    actualState[7] = msg->pose.pose.position.z;
    cout << actualState[0] << " " << actualState[1] << " " << actualState[2] << " " << actualState[3] << endl;
    }
}

void RobotFollower::secondFollowerPositionCallback(const nav_msgs::OdometryConstPtr &msg){
    actualPositions[N_STATES] = msg->pose.pose.position.x;
    actualPositions[N_STATES+1] = msg->pose.pose.position.y;
    actualPositions[N_STATES+3] = fmod(tf::getYaw(msg->pose.pose.orientation)+2*PI, 2*PI);//camera heading
}

void RobotFollower::thirdFollowerPositionCallback(const nav_msgs::OdometryConstPtr &msg){
    actualPositions[2*N_STATES] = msg->pose.pose.position.x;
    actualPositions[2*N_STATES+1] = msg->pose.pose.position.y;
    actualPositions[2*N_STATES+3] = fmod(tf::getYaw(msg->pose.pose.orientation)+2*PI, 2*PI);//camera heading
}

void RobotFollower::secondFollowerTrajectoryCallback(const mbzirc_trackers::TrackerTrajectoryConstPtr &msg){
    mbzirc_trackers::TrackerTrajectory trajectory;
    trajectory.points = msg->points;
    for (int i = 0; i < LENGTH_HORIZONT-N; i++){
        followersTrajectories[i] = trajectory.points[N+i].x;
        followersTrajectories[LENGTH_HORIZONT-N+i] = trajectory.points[N+i].y;
        followersTrajectories[2*(LENGTH_HORIZONT-N)+i] = trajectory.points[N+i].z;
    }
}

void RobotFollower::thirdFollowerTrajectoryCallback(const mbzirc_trackers::TrackerTrajectoryConstPtr &msg){
    mbzirc_trackers::TrackerTrajectory trajectory;
    trajectory.points = msg->points;
    for (int i = 0; i < LENGTH_HORIZONT-N; i++){
        followersTrajectories[(3*LENGTH_HORIZONT-N)+i] = trajectory.points[N+i].x;
        followersTrajectories[(4*LENGTH_HORIZONT-N)+i] = trajectory.points[N+i].y;
        followersTrajectories[(5*LENGTH_HORIZONT-N)+i] = trajectory.points[N+i].z;
    }
}

void RobotFollower::syncNumberCallback(const std_msgs::Int32 &msg){
    counter = msg.data;
}

void RobotFollower::publishDesiredTrajectory(ros::Publisher & pubMessage) const {
    mbzirc_trackers::TrackerTrajectory new_traj;
    mbzirc_trackers::TrackerPoint point;
    for(int i = 0; i < LENGTH_HORIZONT; i++){
        point.x = desiredTrajectory[i];
        point.y = desiredTrajectory[LENGTH_HORIZONT+i];
        point.z = desiredTrajectory[4*LENGTH_HORIZONT+i];//should be set
        point.yaw = desiredTrajectory[3*LENGTH_HORIZONT+i];
        new_traj.points.push_back(point);
    }
    new_traj.header.stamp = ros::Time::now();
    new_traj.header.frame_id = "local_origin";
    new_traj.use_yaw = true;//should be always true, false doesn't make sense
    new_traj.fly_now = true;
    pubMessage.publish(new_traj);
}

void RobotFollower::publishDesiredPitch(ros::Publisher & pubMessage) const {
    mbzirc_trackers::TrackerTrajectory new_traj;
    mbzirc_trackers::TrackerPoint point;
    for(int i = 0; i < LENGTH_HORIZONT; i++){
        point.yaw = desiredTrajectory[5*LENGTH_HORIZONT+i];
        new_traj.points.push_back(point);
    }
    new_traj.header.stamp = ros::Time::now();
    new_traj.header.frame_id = "local_origin";
    new_traj.use_yaw = true;//should be always true, false doesn't make sense
    new_traj.fly_now = true;
    pubMessage.publish(new_traj);
}

void RobotFollower::publishReady(ros::Publisher & pubMessage) const{
    std_msgs::Bool msg;
    msg.data = ready;
    pubMessage.publish(msg);
}

void RobotFollower::flyToTrajectoryStart(ros::Publisher &pubMessage) const{
    mbzirc_trackers::TrackerTrajectory new_traj;
    mbzirc_trackers::TrackerPoint point;
    point.x = desiredTrajectory[0];
    point.y = desiredTrajectory[LENGTH_HORIZONT];
    point.z = desiredTrajectory[4*LENGTH_HORIZONT];
    point.yaw = desiredTrajectory[3*LENGTH_HORIZONT];
    new_traj.points.push_back(point);
    new_traj.header.stamp = ros::Time::now();
    new_traj.header.frame_id = "local_origin";
    new_traj.use_yaw = true;//should be always true
    new_traj.fly_now = true;
    pubMessage.publish(new_traj);
    cout << "Fly to trajectory start" << endl;

}

void RobotFollower::formationStartCallback(const std_msgs::Bool &msg){
    start = msg.data;
}

int main(int argc, char **argv){
    double lumInt, vMax, KMax, verMax, angMax, pitchMax, lightRatio;
    int type, index, nRobots;
    bool finish = false, tmp = false;
    ros::init(argc, argv, "follower_light");
    ROS_INFO("The node started!");
    string addressMap, addressObstacles,stringName;
    ros::NodeHandle privateNodeHandle("~");
    privateNodeHandle.param("name", stringName, string("{0,1,2}"));
    privateNodeHandle.param("addressMap", addressMap, string("cfg/LoadMap"));
    privateNodeHandle.param("addressObstacles", addressObstacles, string("cfg/LoadObstacles")); //default value is nonsense
    privateNodeHandle.param("light_intensity", lumInt, double(4000));
    privateNodeHandle.param("type", type, int(-1));
    privateNodeHandle.param("index", index, int(0));
    privateNodeHandle.param("number_of_robots", nRobots, int(1));
    privateNodeHandle.param("v_max", vMax, double(0.5));
    privateNodeHandle.param("ver_max", verMax, double(0.5));
    privateNodeHandle.param("k_max", KMax, double(2));
    privateNodeHandle.param("ang_max", angMax, double(0.1));
    privateNodeHandle.param("pitch_max", pitchMax, double(0.1));
    privateNodeHandle.param("light_ratio", lightRatio, double(1));
    cout << "index: " << index << endl;
    ros::init(argc, argv, "follower_light"+index);
    ros::NodeHandle nh("~");
    Map map;
    map.loadMap(addressMap.c_str());
    map.loadObstacles(addressObstacles.c_str());
    //map.segmentationMap(TS*VELOCITY_LEADER);
    char numstr[50];
    sprintf(numstr, "%d", index);
    //string temp = "/home/vitkratky/cvut/project/data/";
    //temp=temp+numstr;
    //strcpy(numstr,temp.c_str());
    double *positionOfRobots, *subscribedTrajectory;
    positionOfRobots= (double*)calloc (nRobots*N_STATES,sizeof(double));
    subscribedTrajectory = (double*)calloc (LENGTH_HORIZONT*5,sizeof(double));
    RobotFollower * follower = new RobotFollower(type, lumInt, numstr, positionOfRobots, index, nRobots,
                                                 vMax, KMax, verMax, angMax, pitchMax, lightRatio);
    follower->setPointsOfInterest(map.getPointsOfInterest());
    follower->setObstacles(map.getObstacles());
    follower->subscribedTrajectory = subscribedTrajectory;
    ros::Rate loop_rate(1/(TS*N)); // frequency in Hz
    follower->resetTime();
    while (nh.ok() && !follower->isTrajectorySet) {
        ros::spinOnce();
        loop_rate.sleep();
    }
    follower->setDesiredTrajectory(follower->subscribedTrajectory);
    follower->setInitialState();
    while (nh.ok() && !follower->ready) {
        ros::spinOnce();
        if(follower->distanceFromDesiredPosition() < 1){
            follower->ready = true;
        }
        follower->flyToTrajectoryStart(follower->pubDesiredTrajectory);
        follower->publishReady(follower->pubReady);
        loop_rate.sleep();
    }
    follower->run();
    while (nh.ok() && !finish) {
        ros::spinOnce();
        if(follower->start){
            follower->publishDesiredTrajectory(follower->pubDesiredTrajectory);
            follower->publishDesiredPitch(follower->pubDesiredPitch);
            follower->run();
            tmp = true;
        }else if(tmp){
            finish = true;
        }
        loop_rate.sleep();
    }
    ros::Duration(10);
    return EXIT_SUCCESS;
}

void RobotFollower::reInitialization(void){
    for (int j = 0; j < LENGTH_HORIZONT-N; j++) {
        x[j] = x[j+N];
        x[LENGTH_HORIZONT+j] = x[LENGTH_HORIZONT+j+N];
        x[2*LENGTH_HORIZONT+j] = x[2*LENGTH_HORIZONT+j+N];
        x[3*LENGTH_HORIZONT+j] = x[3*LENGTH_HORIZONT+j+N];
        x[4*LENGTH_HORIZONT+j] = x[4*LENGTH_HORIZONT+j+N];
    }
    for (int j = LENGTH_HORIZONT-N; j < LENGTH_HORIZONT; j++) {
        x[j] = VELOCITY_LEADER;
        x[LENGTH_HORIZONT+j] = 0;
        x[2*LENGTH_HORIZONT+j] = 0;
        x[3*LENGTH_HORIZONT+j] = 0;
        x[4*LENGTH_HORIZONT+j] = 0;
    }
    setDesiredTrajectory(subscribedTrajectory);
}


void RobotFollower::run() {
    runtime.init();
    reInitialization();
    runtime.init();
    Solver * solver = new Solver (actualState, desiredTrajectory, leaderDesiredTrajectory, followersTrajectories,
                                   x, vMax, KMax, verMax, actualPositions, obstacles,
                                   indexOfRobot, nRobots, nObstacle, pointsOfInterest, counter);
    AngularSolver * angSolver = new AngularSolver (actualState, x, angMax, pitchMax,
                                                      pointsOfInterest, counter);
    runtime.now();
    sObj = solver->getObjectiveFunctionValue();
    sObjA = angSolver->getObjectiveFunctionValue();
    //stream << runtime.sec << " " << runtime.nsec << " " << sObj << " " << sObjA << endl;
    computeStates(x);
    nIter = 0;
    delete solver;
    delete angSolver;

    for (int nIter = 0; nIter < N; nIter++) {
    input.forwardVelocity = x[nIter];
    input.curvature = x[LENGTH_HORIZONT+nIter];
    input.angularVelocity = x[3*LENGTH_HORIZONT+nIter];
    input.ascentVelocity = x[2*LENGTH_HORIZONT+nIter];
    input.pitchVelocity = x[4*LENGTH_HORIZONT+nIter];
    move();//shifts the actual state and count actual state for phi - important
    /*states_output << actualState[0] << " " << actualState[1] << " " << " " << actualState[7] << " " << actualState[2] << " " <<
                     actualState[3] << " " << actualState[4] << " " << actualState[5] << " " <<
                     actualState[6] << " " << actualState[8] << " " << actualState[9] << " " << actualState[10] << endl;*/
    counter++;
    cout << "counter: " << counter << endl;
    }
    runtime.now();
}

double RobotFollower::distanceFromDesiredPosition(){
    return sqrt(pow(actualState[0] - desiredTrajectory[0], 2)+pow(actualState[1] -
            desiredTrajectory[LENGTH_HORIZONT], 2)+pow(actualState[7] -
            desiredTrajectory[4*LENGTH_HORIZONT], 2)) + fmod(desiredTrajectory[3*LENGTH_HORIZONT]-actualState[3] +4*PI, 2*PI);
}

void RobotFollower::setDesiredTrajectory(const double * leaderCoo) {
    leaderDesiredTrajectory = leaderCoo;
    double psiFol, dist_xy, v_fx, v_fy, v_fz, v_fxy, alpha, beta;
        for (int j = 0; j < LENGTH_HORIZONT; j++) {
            cout << j << endl;
            v_fx = pointsOfInterest[counter+j].x-leaderCoo[j];

            v_fy = pointsOfInterest[counter+j].y-leaderCoo[LENGTH_HORIZONT+j];
            v_fz = pointsOfInterest[counter+j].z-leaderCoo[3*LENGTH_HORIZONT+j];
            v_fxy = sqrt(pow(v_fx, 2)+pow(v_fy, 2));
            alpha = fabs(fmod(atan2(v_fy, v_fx)+2*PI, 2*PI)-leaderCoo[2*LENGTH_HORIZONT+j])-CAMERA_ANGLE/2;
            beta = fabs(fmod(atan2(v_fz, v_fxy)+2*PI, 2*PI)-leaderCoo[4*LENGTH_HORIZONT+j])-CAMERA_ANGLE_Z/2;
            if(alpha < 0){
                psiFol = fmod(leaderCoo[2*LENGTH_HORIZONT+j] + pointsOfInterest[counter+j].psi +2*PI, 2*PI);//previous
            }else{
                psiFol = fmod(atan2(pointsOfInterest[counter+j].y-leaderCoo[LENGTH_HORIZONT+j],
                        pointsOfInterest[counter+j].x-leaderCoo[j])+pointsOfInterest[counter+j].psi +2*PI, 2*PI);
            }
            if(beta < 0){
                desiredTrajectory[5*LENGTH_HORIZONT+j] = leaderCoo[4*LENGTH_HORIZONT+j]-pointsOfInterest[counter+j].eta; //previous
            }else{
                desiredTrajectory[5*LENGTH_HORIZONT+j] = fabs(fmod(atan2(v_fz, v_fxy)+2*PI, 2*PI))-pointsOfInterest[counter+j].eta; //previous
            }
            desiredTrajectory[j] = pointsOfInterest[counter+j].x-dist*cos(psiFol);
            desiredTrajectory[LENGTH_HORIZONT+j] = pointsOfInterest[counter+j].y-dist*sin(psiFol);
            desiredTrajectory[2*LENGTH_HORIZONT+j] = 0;
            desiredTrajectory[3*LENGTH_HORIZONT+j] = fmod(atan2(pointsOfInterest[counter+j].y -desiredTrajectory[LENGTH_HORIZONT+j],
                                pointsOfInterest[counter+j].x - desiredTrajectory[j])+2*PI, 2*PI);
            dist_xy = sqrt(pow(desiredTrajectory[j]-pointsOfInterest[counter+j].x,2)
                    +pow(desiredTrajectory[LENGTH_HORIZONT+j]-pointsOfInterest[counter+j].y,2));
            desiredTrajectory[4*LENGTH_HORIZONT+j] = pointsOfInterest[counter+j].z + dist_xy*tan(leaderCoo[4*LENGTH_HORIZONT+j]+pointsOfInterest[counter+j].eta);
        }
}

void RobotFollower::setInitialState() {
    nIter = N;
    actualState[2] = fmod(atan2(desiredTrajectory[LENGTH_HORIZONT+1]-desiredTrajectory[LENGTH_HORIZONT],
                desiredTrajectory[1]-desiredTrajectory[0]) +2*PI, 2*PI);
    actualState[4] = input.forwardVelocity;
    actualState[5] = input.curvature;
    actualState[6] = input.angularVelocity;
    actualState[7] = pointsOfInterest[0].z;
    actualState[8] = input.ascentVelocity;
}

void RobotFollower::setPointsOfInterest(vCoordinates & iCameraPoints) {
    pointsOfInterest = iCameraPoints;
}

void RobotFollower::setState(double *x, double stateX, double stateY, double stateZ, double statePhi, double statePsi, double stateEta, int j){
    if(fabs(x[LENGTH_HORIZONT+j]) < (1E-6F)){ //if K==0
        desiredTrajectory[j]= stateX + x[j]*cos(statePhi)*TS;
        desiredTrajectory[LENGTH_HORIZONT+j]= stateY + x[j]*sin(statePhi)*TS;
    }else{
        desiredTrajectory[j]= stateX + (-sin(statePhi)+sin(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
        desiredTrajectory[LENGTH_HORIZONT+j]= stateY - (-cos(statePhi)+cos(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
    }
    desiredTrajectory[2*LENGTH_HORIZONT+j] = statePhi + x[LENGTH_HORIZONT+j]*x[j]*TS;
    desiredTrajectory[3*LENGTH_HORIZONT+j] = fmod(statePsi + x[3*LENGTH_HORIZONT+j]*TS + 2*PI, 2*PI);
    desiredTrajectory[4*LENGTH_HORIZONT+j] = stateZ + x[2*LENGTH_HORIZONT+j]*TS;
    desiredTrajectory[5*LENGTH_HORIZONT+j] = stateEta + x[4*LENGTH_HORIZONT+j]*TS;
}

void RobotFollower::computeStates(double *x){
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        if(j == 0){
            setState(x, actualState[0], actualState[1], actualState[7], actualState[2], actualState[3], actualState[9], j);
        }else{
            setState(x, desiredTrajectory[j-1], desiredTrajectory[LENGTH_HORIZONT+j-1],
                desiredTrajectory[4*LENGTH_HORIZONT+j-1], desiredTrajectory[2*LENGTH_HORIZONT+j-1],
                desiredTrajectory[3*LENGTH_HORIZONT+j-1], desiredTrajectory[5*LENGTH_HORIZONT+j-1], j);
        }
    }
}

