/**
    solver.cc
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#include <iostream>
#include "solver.h"
extern "C" {
int x_is_new;
double objrep,objeps,gLgeps;
#include "cfsqpusr.h"
}

using namespace std;

Solver::Solver(const double *actualState, const double *desiredTrajectory, const double *leaderDesiredTrajectory,
               const double *secondFollowerTrajectory, double *x, const double vmax, const double Kmax, const double vertMax,
               double *actualPositions, double *obstacles, int index, int nRobots, int nObstacles, vCoordinates poi, int counter)
    : actualState(actualState), desiredTrajectory(desiredTrajectory), x(x), vMax(vmax), KMax(Kmax), angMax(angMax), verMax(vertMax),
        leaderDesiredTrajectory(leaderDesiredTrajectory), counter(counter), otherFollowersTrajectories(secondFollowerTrajectory){
    states = (double *)calloc(4*LENGTH_HORIZONT,sizeof(double));
    this->index = index;
    this->nRobots= nRobots;
    this->actualPositions = actualPositions;
    this->obstacles = obstacles;
    this->nObstacles = nObstacles;
    this->poi = poi;
    callSolver();
}

Solver::~Solver() {
    free(states);
}

void Solver::generateBoundaryValues(double * bl,double * bu){
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        bl[j] = -vMax;
        bl[LENGTH_HORIZONT+j] = -KMax;
        bl[2*LENGTH_HORIZONT+j] = -verMax;
        bu[j] = vMax;
        bu[LENGTH_HORIZONT+j] = KMax;
        bu[2*LENGTH_HORIZONT+j] = verMax;
    }
}

void  Solver::setState(double *x, double stateX, double stateY, double stateZ, double statePhi, int j){
    if (fabs(x[LENGTH_HORIZONT+j]) < (1E-6F)) { //if K==0
        states[j]= stateX + x[j]*cos(statePhi)*TS;
        states[LENGTH_HORIZONT+j]= stateY + x[j]*sin(statePhi)*TS;
    } else {
        states[j]= stateX + (-sin(statePhi)+sin(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
        states[LENGTH_HORIZONT+j]= stateY - (-cos(statePhi)+cos(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
    }
    states[2*LENGTH_HORIZONT+j]= statePhi + x[LENGTH_HORIZONT+j]*x[j]*TS;
    states[3*LENGTH_HORIZONT+j] = stateZ + x[2*LENGTH_HORIZONT+j]*TS;
}

void Solver::computeStates(double *x){
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        if (j == 0){
            setState(x, actualState[0], actualState[1], actualState[7], actualState[2], j);
        } else {
            setState(x, states[j-1], states[LENGTH_HORIZONT+j-1], states[3*LENGTH_HORIZONT+j-1],
                states[2*LENGTH_HORIZONT+j-1], j);
        }
    }
}

double Solver::getObjectiveFunctionValue(){
    return objf;
}


void objF(int nparam, int j,double *x,double *fj,void *cd) {
    Solver *so = (Solver *) cd;
    so->computeStates(x);
    double posDev=0, vDiff=0, KDiff=0, wDiff=0,fovDev=0, collFoll=0, ds, dy, dx, diffAlpha, alpha, tmp1, dz, beta, diffBeta, ds1, ds2;
    vDiff+=pow(so->actualState[4]-x[0], 2);
    KDiff+=pow(so->actualState[5]-x[LENGTH_HORIZONT], 2);
    wDiff+=(so->actualState[8]-x[2*LENGTH_HORIZONT], 2);
    for (int i = 1; i < LENGTH_HORIZONT; ++i) {
        vDiff+=pow(x[i-1]-x[i], 2);
        KDiff+=pow(x[LENGTH_HORIZONT+i-1]-x[LENGTH_HORIZONT+i], 2);
        wDiff+=pow(x[2*LENGTH_HORIZONT+i-1]-x[2*LENGTH_HORIZONT+i], 2);
    }

    for (int i = 0; i < LENGTH_HORIZONT; i++) {
        posDev+=pow(so->states[i] - so->desiredTrajectory[i],2) + pow(so->states[LENGTH_HORIZONT+i] - so->desiredTrajectory[LENGTH_HORIZONT+i],2)
                +pow(so->states[3*LENGTH_HORIZONT+i]- so->desiredTrajectory[4*LENGTH_HORIZONT+i], 2);
        dy = so->states[LENGTH_HORIZONT+i] - so->leaderDesiredTrajectory[LENGTH_HORIZONT+i];
        dx = so->states[i]-so->leaderDesiredTrajectory[i];
        dz = so->states[3*LENGTH_HORIZONT+i]-so->leaderDesiredTrajectory[3*LENGTH_HORIZONT+i];
        alpha = fabs(fmod(atan2(dy, dx) + 2*PI, 2*PI)-so->leaderDesiredTrajectory[2*LENGTH_HORIZONT+i]);
        beta = fabs(fmod(atan2(dz, sqrt(pow(dx, 2)+pow(dy, 2)))+2*PI, 2*PI)-so->leaderDesiredTrajectory[4*LENGTH_HORIZONT+i]);
        diffAlpha = alpha > PI ? 2*PI-alpha : alpha;
        diffBeta = beta > PI ? 2*PI-beta : beta;
        ds1 = sqrt(pow(dx, 2)+pow(dy, 2))*sin(diffAlpha-CAMERA_ANGLE/2);
        ds2 = sqrt(pow(dy, 2)+pow(dz, 2))*sin(diffBeta-CAMERA_ANGLE_Z/2);
        tmp1 = diffAlpha > PI/2 || diffBeta > PI/2 ? 0 : 1;
        ds1 = ds1 > 0 ? ds1 : 0;
        ds2 = ds2 > 0 ? ds2 : 0;
        ds = sqrt(pow(ds1, 2) + pow(ds2, 2));
        if(ds > DIST_A){
            fovDev += (ds - DIST_D)/(ds-DIST_A) < 0 ? tmp1*pow((ds-DIST_D)/(ds-DIST_A), 2) : 0;
        }
    }

    double xp, yp, zp, xo, yo, d;
    double coll=0, dist=0, actDist, obscoll=0, type, obsLen, projLen, min, df;

    if(so->nRobots != 0 ){
        for (int k = 0; k < LENGTH_HORIZONT; k++) {
            min=DBL_MAX;
            for (int i = 0; i < so->nRobots-1; i++) {
                    xp = double(so->actualState[i*N_STATES]);
                     yp = double(so->actualState[i*N_STATES+1]);
                    zp = double(so->actualState[i*N_STATES+7]);
                    dist=sqrt(pow(so->states[k] - xp,2) + pow(so->states[LENGTH_HORIZONT+k] -  yp,2)+
                    pow(so->states[3*LENGTH_HORIZONT+k] - zp,2))-ROBOT_RADIUS;
                if (dist<min) {
                    min=dist;
                }
            }
            coll+=(min - RADIUS_D_ROB)/(min-RADIUS_A_ROB) < 0 ? pow((min - RADIUS_D_ROB)/(min-RADIUS_A_ROB), 2) : 0;
        }

        for(int k=LENGTH_HORIZONT-1; k >-1; k--){
            df = RADIUS_D_ROB;
            d = sqrt(pow(so->states[k]-so->leaderDesiredTrajectory[k],2)+
                     pow(so->states[LENGTH_HORIZONT+k]-so->leaderDesiredTrajectory[LENGTH_HORIZONT+k],2)+
                    pow(so->states[3*LENGTH_HORIZONT+k]-so->leaderDesiredTrajectory[3*LENGTH_HORIZONT+k],2))-
                    ROBOT_RADIUS;
            if(k<(LENGTH_HORIZONT-N)){
                for (int i = 0; i < so->index; ++i) {
                    df = sqrt(pow(so->states[k]-so->otherFollowersTrajectories[i*(LENGTH_HORIZONT*3)+k],2)+
                        pow(so->states[LENGTH_HORIZONT+k]-so->otherFollowersTrajectories[i*(LENGTH_HORIZONT*3)+LENGTH_HORIZONT-N+k],2)+
                        pow(so->states[3*LENGTH_HORIZONT+k]-so->otherFollowersTrajectories[i*(LENGTH_HORIZONT*3)+2*(LENGTH_HORIZONT-N)+k],2))-
                        ROBOT_RADIUS;
                    d = df < d ? df : d;
                }
            }
            collFoll+= (d - RADIUS_D_ROB)/(d-RADIUS_A_ROB) < 0 ? pow((d - RADIUS_D_ROB)/(d-RADIUS_A_ROB), 2) : 0;
        }
    }

    if(so->nObstacles != 0){ //collision of robot with obstacles
        for (int k = 0; k < LENGTH_HORIZONT; k++) {
            min = DBL_MAX;
            for (int i = 0; i < so->nObstacles; i++) {
                xo =double(so->obstacles[6*i+1]);
                yo = double(so->obstacles[6*i+2]);
                d = double(so->obstacles[6*i+5]);
                type = double(so->obstacles[6*i]);
                    if(type < 2){ //circle type
                        actDist=sqrt(pow(so->states[k] - xo,2) + pow(so->states[LENGTH_HORIZONT+k] - yo,2)) - d;
                    }else{ //surface type
                        dx = double(so->obstacles[6*i+3]) - xo;
                        dy = double(so->obstacles[6*i+4]) - yo;
                        obsLen = (dx*dx)+(dy*dy);
                        projLen = ((so->states[k]-xo)*dx +(so->states[LENGTH_HORIZONT+k]-yo)*dy)/obsLen;
                        if(projLen>1){
                            projLen=1;
                        }else if(projLen<0){
                            projLen=0;
                        }
                        dx = xo + projLen*dx - so->states[k];
                        dy = yo + projLen*dy - so->states[LENGTH_HORIZONT+k];
                        actDist = sqrt(dx*dx + dy*dy)-d;
                    }
                    if (actDist<min){
                        min=actDist;
                    }
                }
                obscoll += (min - RADIUS_D)/(min-RADIUS_A) < 0 ? k*pow((min - RADIUS_D)/(min-RADIUS_A), 2) : 0;
       }
     }
    *fj=(1*posDev+0.2*vDiff+0.2*KDiff+fovDev*2+0.4*wDiff+2*obscoll+N*collFoll+coll)/LENGTH_HORIZONT;
    return;
}

void cntrF(int nparam, int j, double *x, double *gj, void *cd) {
    Solver *so = (Solver *) cd;
    so->computeStates(x);
    double minDist=DBL_MAX, z, vFx, vFy, vFz, vFxy, alpha, diffAlpha, angDist, angDist2, diffBeta, beta;

    for (int i = 0; i < LENGTH_HORIZONT; i++) {
        z = so->states[3*LENGTH_HORIZONT+i];
        if(z < Z_MIN|| z > Z_MAX){
            minDist = 0;
            break;
        }
        vFx = so->states[i]-so->leaderDesiredTrajectory[i];
        vFy = so->states[LENGTH_HORIZONT+i]-so->leaderDesiredTrajectory[LENGTH_HORIZONT+i];
        vFz = so->states[3*LENGTH_HORIZONT+i]-so->leaderDesiredTrajectory[3*LENGTH_HORIZONT+i];
        vFxy = sqrt(pow(vFx, 2)+pow(vFy, 2));
        alpha = fabs(fmod(atan2(vFy, vFx)+2*PI, 2*PI)-so->leaderDesiredTrajectory[2*LENGTH_HORIZONT+i]);
        beta = fabs(fmod(atan2(vFz, vFxy)+2*PI, 2*PI)-so->leaderDesiredTrajectory[4*LENGTH_HORIZONT+i]);
        diffBeta = beta > PI ? 2*PI-beta : beta;
        diffAlpha = alpha > PI ? 2*PI-alpha : alpha;
        angDist = diffAlpha - CAMERA_ANGLE/2;
        angDist2 = diffBeta - CAMERA_ANGLE_Z/2;
        if(angDist < 0 && angDist2 < 0){
            minDist = angDist;
            break;
        }
    }
    if(!(minDist < RADIUS_A)){
        if(so->nRobots != 0 ){ //collision of follower with other robots
            double xp, yp, zp, dist, d, dl, df;
            for (int i = 0; i < so->nRobots-1; i++) {
                xp = double(so->actualPositions[i*11]);
                yp = double(so->actualPositions[i*11+1]);
                zp = double(so->actualPositions[i*11+7]);
                for (int k = 0; k < LENGTH_HORIZONT; k++) {
                    dist=sqrt(pow(so->states[k] - xp,2) + pow(so->states[LENGTH_HORIZONT+k] - yp,2)+
                            pow(so->states[3*LENGTH_HORIZONT+k] - zp,2))-ROBOT_RADIUS;
                    if (dist<RADIUS_A){
                        minDist=dist;
                        break;
                    }
                }
            }
            for(int k=LENGTH_HORIZONT-1; k >-1; k--){
                df = RADIUS_D;
                dl = sqrt(pow(so->states[k]-so->leaderDesiredTrajectory[k],2)+
                          pow(so->states[LENGTH_HORIZONT+k]-so->leaderDesiredTrajectory[LENGTH_HORIZONT+k],2)+
                         pow(so->states[3*LENGTH_HORIZONT+k]-so->leaderDesiredTrajectory[3*LENGTH_HORIZONT+k],2))-
                         ROBOT_RADIUS;
                if(k<(LENGTH_HORIZONT-N)){
                    for (int i = 0; i < so->index; ++i) {
                        df = sqrt(pow(so->states[k]-so->otherFollowersTrajectories[i*(LENGTH_HORIZONT*3)+k],2)+
                            pow(so->states[LENGTH_HORIZONT+k]-so->otherFollowersTrajectories[i*(LENGTH_HORIZONT*3)+LENGTH_HORIZONT-N+k],2)+
                            pow(so->states[3*LENGTH_HORIZONT+k]-so->otherFollowersTrajectories[i*(LENGTH_HORIZONT*3)+2*(LENGTH_HORIZONT-N)+k],2))-
                            ROBOT_RADIUS;
                        d = df < d ? df : d;
                    }
                }
                d = df < dl ? df : dl;
                if(d < RADIUS_A){
                    minDist = d;
                    break;
                }
            }
        }
    }
    if(!(minDist < RADIUS_A)){
        if(so->nObstacles != 0){ //collision of robot with obstacles
            double type, d, xo, yo, obsLen, projLen, dist, dx, dy;
            minDist = DBL_MAX;
            for (int i = 0; i < so->nObstacles; i++) {
                xo =double(so->obstacles[6*i+1]);
                yo = double(so->obstacles[6*i+2]);
                d = double(so->obstacles[6*i+5]);
                type = double(so->obstacles[6*i]);
                for (int k = 0; k < LENGTH_HORIZONT; k++) {
                    if(type < 2){ //circle type
                        dist=sqrt(pow(so->states[k] - xo,2) + pow(so->states[LENGTH_HORIZONT+k] - yo,2)) - d;
                    }else{ //surface type
                        dx = double(so->obstacles[6*i+3]) - xo;
                        dy = double(so->obstacles[6*i+4]) - yo;
                        obsLen = (dx*dx)+(dy*dy);
                        projLen = ((so->states[k]-xo)*dx +(so->states[LENGTH_HORIZONT+k]-yo)*dy)/obsLen;
                        if(projLen>1){
                            projLen=1;
                        }else if(projLen<0){
                            projLen=0;
                        }
                        dx = xo + projLen*dx - so->states[k];
                        dy = yo + projLen*dy - so->states[LENGTH_HORIZONT+k];
                        dist = sqrt(dx*dx + dy*dy)-d;
                    }
                    if (dist<RADIUS_A){
                        minDist=dist;
                        break;
                    }
                }
            }
        }
    }
    *gj = RADIUS_A-minDist;
    return;
}

void Solver::callSolver (void) {
    int nparam,nf,nineq,neq,mode,iprint,miter,neqn,nineqn,
            ncsrl,ncsrn,nfsr,mesh_pts[1],inform;
    double bigbnd,eps,epsneq,udelta;
    double *bl,*bu,*f,*g,*lambda,*cd;
    mode=100;
    iprint=0; //printing informations
    miter=500; //maximum number of iterations
    bigbnd=1.e12; // all bounds above bigbnd or under -bigbnd in bu or bl are suposed to be infinite
    eps=1.e-9; //final norm requirement
    epsneq=1.e-6; // maximum allowed violation of nonlinear equality constraints
    udelta=0.e0; //perturbation size
    nparam=(N_PARAMS_F-2)*LENGTH_HORIZONT; //dimension of x
    nf=1; //number of objective functions
    neqn=neq=0; //3; //nonlinear equality constraints / equality constraints
    nineqn=1; //nonlinear inequality constraints / inequality constraints
    nineq=1;
    ncsrn=0; //sets of nonlinear sequentially related constraints
    ncsrl=0; //sets of linear sequentially related constraints
    mesh_pts[1]=0;
    nfsr=0; //sequentially related objective functions
    cd=(double *)calloc(nparam,sizeof(double));
    bl=(double *)calloc(nparam,sizeof(double)); //lower bounds
    bu=(double *)calloc(nparam,sizeof(double)); // upper bound
    f=(double *)calloc(nf+1,sizeof(double));
    g=(double *)calloc(nineq+(mesh_pts[1]-1)*(ncsrl+ncsrn)+neq+1,sizeof(double));
    lambda=(double *)calloc(nineq+(mesh_pts[1]-1)*(ncsrl+ncsrn)+neq+nf+nparam,sizeof(double));
    generateBoundaryValues(bl,bu);
    cfsqp(nparam,nf,nfsr,nineqn,nineq,neqn,neq,ncsrl,ncsrn,mesh_pts,
          mode,iprint,miter,&inform,bigbnd,eps,epsneq,udelta,bl,bu,x,
          f,g,lambda,objF,cntrF,grobfd,grcnfd,this);
    objF(5,5,x,&objf,this);
    free(bl);
    free(bu);
    free(f);
    free(g);
    free(lambda);
    free(cd);
}
