/**
    angularsolver.h

    @author Vit Kratky
    @version 1.3 20/05/17
*/

#ifndef ANGULARSOLVER_H
#define ANGULARSOLVER_H

#include <cstdlib>
#include <iostream>
#include "robot.h"
#include "leader.h"

using namespace std;
using namespace mkr::formation;

/**
 * @brief Class AngularSolver provides all variables and functions essential for the optimization task
 *        for camera heading
 */
class AngularSolver {
private:
    const double *actualState;
    double *x;
    const double angMax, pitchMax;
    friend void objA(int nparam,int j, double *x, double *fj, void *cd);
    friend void constrA(int nparam, int j, double *x, double *gj, void *cd);
    double objf;
    void computeStatesPosition(double *x);
    void setState(double *x, double stateX, double stateY, double stateZ,  double statePhi, int j);
public:

    /**
     * @brief constructor
     *
     * @param actualState - the actual state of the leader
     * @param x - vector of control variables
     * @param angMax - maximal absolute value of angular velocity omega_L
     * @param pitchMax - maximal absolute value of angular velocity epsilon_L
     * @param poi - objects of interest
     */
    AngularSolver(const double *actualState, double *x, const double angMax, const double pitchMax, vCoordinates poi);

    /**
     * @brief destructor
     */
    ~AngularSolver();

    /**
     * @brief calls the solver
     */
    void callSolver(void);

    /**
     * @brief generates the boundary values of optimization vector
     *
     * @param bl - lower boundaries
     * @param bu - upper boundaries
     */
    void generateBoundaryValues(double *bl,double *bu);

    /**
     * @brief computes states of the leader within the optimization process
     *
     * @param x - the vector of control variables
     */
    void computeStates(double *x);

    /**
     * @brief returns the value of the objective function corresponding with the resulting vector of decision variables
     */
    double getObjectiveFunctionValue();

    /**
     * @brief states of the leader computed from the vector of control variables for entire length of control horizon
     */
    double *states;
    /**
     * @brief objects of interest
     */
    vCoordinates poi;
};

#endif // ANGULARSOLVER_H
