/**
    controls.h
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/

#ifndef __CONTROLS_H__
#define __CONTROLS_H__

namespace mkr { namespace formation {

/**
 * @brief Class Controls provides an object which collects all control inputs of the robot
 */
class Controls {
public:

    /**
     * @brief constructor
     *
     * @param forwardVelocity - the initial forward velocity
     * @param ascentVelocity - the initial ascent velocity
     * @param curvature - the initial curvature
     * @param angularVelocity - the initial angular velocity omega_L
     * @param pitchVelocity - the initial angular velocity epsilon_L
     */
    Controls(const double forwardVelocity, const double curvature, const double angularVelocity,
             const double ascentVelocity, const double pitchVelocity);

    /**
     * @brief constructor which sets all initial control inputs to 0
     */
    Controls(void);

    /**
     * @brief destructor
     */
    ~Controls(void);

    double forwardVelocity;
    double curvature;
    double angularVelocity;
    double ascentVelocity;
    double pitchVelocity;
};

}}

#endif

