/**
    map.h
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#ifndef __MAP_H__
#define __MAP_H__

#include <vector>

#include "coordinates.h"
#include "obstacle.h"
#include "state.h"
#include "robot.h"


namespace mkr { namespace formation {

typedef std::vector<Obstacle> Obstacles;
typedef std::vector<Coordinates> VCoordinates;
typedef std::vector<int> Modes;

/**
 * @brief Class Map provides the loading of the map and obstacles
 */
class Map {
public:

    /**
     * @brief constructor
     */
	Map(void);

    /**
     * @brief destructor
     */
	~Map(void);

    /**
     * @brief loads the obstacles from the file
     *        expected format: type x_1 y_1 x_2 y_2 radius
     */
	void loadObstacles(const char * name);

    /**
     * @brief loads the map from the file
     *        expected format: x y z OoI_x OoI_y OoI_z mode
     */
	void loadMap(const char * name);

    /**
     * @brief returns the uploaded trajectory
     */
    VCoordinates & getCoordinates(void);

    /**
     * @brief returns the uploaded objects of interest
     */
    VCoordinates & getPointsOfInterest(void);

    /**
     * @brief returns the uploaded obstacles
     */
    Obstacles & getObstacles(void);

private:
	Obstacles obstacles;
	VCoordinates coordinates;
    VCoordinates pointsOfInterest;
    Modes modes;
};

}}

#endif

