/**
    robot.h
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#ifndef __ROBOT_H__
#define __ROBOT_H__

#include <ros/ros.h>
#include <nav_msgs/Odometry.h>

#include <vector>

#include "coordinates.h"
#include "controls.h"
#include "formtime.h"
#include "state.h"
#include "obstacle.h"

#define PI 3.14159265358979323846
#define VELOCITY_LEADER 0.2
#define TS 0.4
#define N 1
#define LENGTH_HORIZONT 13
#define N_PARAMS 5
#define N_STATES 11
#define RADIUS_D 2.5
#define RADIUS_D_ROB 2.5
#define RADIUS_A_ROB 1.5
#define RADIUS_A 1.5
#define Z_MIN 0.5
#define Z_MAX 10
#define ROBOT_RADIUS 0.25
#define ANG_INPUTS_SHIFT 3

namespace mkr { namespace formation {

typedef std::vector<Coordinates> vCoordinates;
typedef std::vector<Obstacle> Obstacles;

/**
 * @brief Class Robot represents the main class for control of the robot
 */
class Robot {
public:

    /**
     * @brief constructor
     */
    Robot();

    /**
     * @brief destructor
     */
    virtual ~Robot();

    /**
     * @brief runs the optimization process and the preparations for it
     */
    virtual void run(void) = 0;

    /**
     * @brief sets the initial state of the leader
     *
     * @param iState - initial state of the robot
     * @param psi - initial psi of the robot
     */
    virtual void setInitialState(Coordinates iState, double psi) = 0;

    /**
     * @brief sets the obstacles
     *
     * @param obstacles - the set of obstacleszzzz
     */z
    void setObstacles(Obstacles obstacles);

    /**
     * @brief resets the time for the measurment of the time of the optimization
     */
    static void resetTime(void);

    /**
     * @brief returns the actual state
     */
    double *getActualState(void);

    /**
     * @brief simulates the motion of the robot
     */
    void move(void);
	
protected:
    int nIter, nObstacle;
    static Time time;
    static Time runtime;
    double *x, *desiredTrajectory, *actualState, *obstacles;
	Controls input;
};
}}

#endif

