/**
    leadersolver.cc

    @author Vit Kratky
    @version 1.3 20/05/17
*/
#include <iostream>
#include "leadersolver.h"
extern "C" {
int x_is_new;
double objrep,objeps,gLgeps;
#include "cfsqpusr.h"
}

using namespace std;

LeaderSolver::LeaderSolver(const double *actualState, const double *desiredTrajectory,
               double *x, const double Kmax , const double Vmax, const double Vertmax, double *obstacles,
               int nRobots, int nObstacles, vCoordinates poi)
    : actualState(actualState), desiredTrajectory(desiredTrajectory), vMax(Vmax), x(x), KMax(Kmax), verMax(Vertmax) {
    states = (double *)calloc(4*LENGTH_HORIZONT,sizeof(double));
    this->nRobots= nRobots;
    this->obstacles = obstacles;
    this->nObstacles = nObstacles;
    this->poi = poi;
    callSolver();
}

LeaderSolver::~LeaderSolver() {
    free(states);
}

void LeaderSolver::generateBoundaryValues(double * bl,double * bu){
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        bl[j] = -vMax;
        bl[LENGTH_HORIZONT+j] = -KMax;
        bl[2*LENGTH_HORIZONT+j] = -verMax;
        bu[j] = vMax;
        bu[LENGTH_HORIZONT+j] = KMax;
        bu[2*LENGTH_HORIZONT+j] = verMax;
    }
}

void  LeaderSolver::setState(double *x, double stateX, double stateY, double stateZ, double statePhi, int j){
    if (fabs(x[LENGTH_HORIZONT+j]) < (1E-6F)) { //if K==0
        states[j]= stateX + x[j]*cos(statePhi)*TS;
        states[LENGTH_HORIZONT+j]= stateY + x[j]*sin(statePhi)*TS;
    } else {
        states[j]= stateX + (-sin(statePhi)+sin(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) /  x[LENGTH_HORIZONT+j];
        states[LENGTH_HORIZONT+j]= stateY - (-cos(statePhi)+cos(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
    }
    states[2*LENGTH_HORIZONT+j]= statePhi + x[LENGTH_HORIZONT+j]*x[j]*TS;
    states[3*LENGTH_HORIZONT+j] = stateZ + x[2*LENGTH_HORIZONT+j]*TS;
}

void LeaderSolver::computeStates(double *x){
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        if (j == 0) setState(x, actualState[0], actualState[1], actualState[7], actualState[2], j);
        else setState(x, states[j-1], states[LENGTH_HORIZONT+j-1], states[3*LENGTH_HORIZONT+j-1], states[2*LENGTH_HORIZONT+j-1], j);
    }
}

double LeaderSolver::getObjectiveFunctionValue(){
    return objf;
}

void obj(int nparam, int j,double *x,double *fj,void *cd) {
    LeaderSolver *so = (LeaderSolver *) cd;
    so->computeStates(x);
    double posDev=0, vDiff=0, KDiff=0, wDiff=0, distDev=0, dist=0;
    KDiff+=pow(so->actualState[5]-x[LENGTH_HORIZONT], 2);
    vDiff+=pow(so->actualState[4]-x[0], 2);
    wDiff+=(so->actualState[8]-x[2*LENGTH_HORIZONT], 2);
    for (int i = 1; i < LENGTH_HORIZONT; ++i) {
        KDiff+=pow(x[LENGTH_HORIZONT+i-1]-x[LENGTH_HORIZONT+i], 2);
        vDiff+=pow(x[i-1]-x[i], 2);
        wDiff+=pow(x[2*LENGTH_HORIZONT+i-1]-x[2*LENGTH_HORIZONT+i], 2);
    }
    for (int k = 0; k < LENGTH_HORIZONT; k++) {
        posDev+=pow(so->states[k] - so->desiredTrajectory[k],2) + pow(so->states[LENGTH_HORIZONT+k] - so->desiredTrajectory[LENGTH_HORIZONT+k],2)
                +pow(so->states[3*LENGTH_HORIZONT+k] - so->desiredTrajectory[4*LENGTH_HORIZONT+k],2);
        dist = sqrt(pow(so->poi[k].x-so->desiredTrajectory[k], 2)+pow(so->poi[k].y-so->desiredTrajectory[LENGTH_HORIZONT+k], 2))-
                sqrt(pow(so->poi[k].x-so->states[k], 2)+pow(so->poi[k].y-so->states[LENGTH_HORIZONT+k], 2));
        distDev += dist < 0 ? pow(dist, 2) : 0;
    }

    double min, coll=0, obscoll=0, xp, yp, zp, xo, yo, d;
    if(so->nRobots != 0 ){ //collision of leader with followers
        for (int k = 0; k < LENGTH_HORIZONT; k++) {
            min=DBL_MAX;
            for (int i = 1; i < so->nRobots; i++) {
                xp = double(so->actualState[i*N_STATES]);
                yp = double(so->actualState[i*N_STATES+1]);
                zp = double(so->actualState[i*N_STATES+7]);
                dist=sqrt(pow(so->states[k] - xp,2) + pow(so->states[LENGTH_HORIZONT+k] - yp,2)+
                pow(so->states[3*LENGTH_HORIZONT+k] - zp,2))-ROBOT_RADIUS;
                if (dist<min) {
                    min=dist;
                }
            }
            coll+=(min - RADIUS_D)/(min-RADIUS_A) < 0 ? pow((min - RADIUS_D)/(min-RADIUS_A), 2) : 0;
        }
    }

    double actDist, dx, dy, type, x2, y2, obsLen, projLen, xProj, yProj;
    if(so->nObstacles != 0){ //collision of robot with obstacles
          for (int k = 0; k < LENGTH_HORIZONT; k++) {
              min = DBL_MAX;
              for (int i = 0; i < so->nObstacles; i++) {
                  xo = double(so->obstacles[6*i+1]);
                  yo = double(so->obstacles[6*i+2]);
                  d = double(so->obstacles[6*i+5]);
                  type = double(so->obstacles[6*i]);
                  if(type < 2){ //circle type
                     actDist=sqrt(pow(so->states[k] - xo,2) + pow(so->states[LENGTH_HORIZONT+k] - yo,2)) - d;
                  }else{ //surface type
                     x2 = double(so->obstacles[6*i+3]);
                     y2 = double(so->obstacles[6*i+4]);
                     dx = x2 - xo;
                     dy = y2 - yo;
                     obsLen = (dx*dx)+(dy*dy);
                     projLen = ((so->states[k]-xo)*dx +(so->states[LENGTH_HORIZONT+k]-yo)*dy)/(obsLen);
                     if(projLen>1){
                        projLen=1;
                     }else if(projLen<0){
                        projLen=0;
                     }
                     dx = xo + projLen*dx + xProj - so->states[k];
                     dy = yo + projLen*dy + yProj - so->states[LENGTH_HORIZONT+k];
                     actDist = sqrt(dx*dx + dy*dy)-d;
                  }
                  if (actDist < min){
                      min=actDist;
                  }
              }
              obscoll += (min - RADIUS_D)/(min-RADIUS_A) < 0 ? k*pow((min - RADIUS_D)/(min-RADIUS_A), 2) : 0;
          }
    }
    *fj=(1*posDev+0.1*wDiff+0.1*vDiff+0.1*KDiff+0.5*distDev+0.5*obscoll+0.2*coll)/LENGTH_HORIZONT;
    return;
}

void cntr(int nparam, int j, double *x, double *gj, void *cd) {
    LeaderSolver *so = (LeaderSolver *) cd;
    so->computeStates(x);
    double minDist=N_STATES;
    double dist, xp, yp, zp, z=0;
    for (int i = 0; i < LENGTH_HORIZONT; i++) {
        z = so->states[3*LENGTH_HORIZONT+i];
        if(z < Z_MIN|| z > Z_MAX){
            minDist = 0;
            break;
        }
    }
    if(!(minDist < RADIUS_A)){
        if(so->nRobots != 0 ){ //collision of leader with followers
            for (int i = 1; i < so->nRobots; i++) {
                xp = double(so->actualState[i*N_STATES]);
                yp = double(so->actualState[i*N_STATES+1]);
                zp = double(so->actualState[i*N_STATES+7]);
                for (int k = 0; k < LENGTH_HORIZONT; k++) {
                    dist=sqrt(pow(so->states[k] - xp,2) + pow(so->states[LENGTH_HORIZONT+k] - yp,2)+
                            pow(so->states[3*LENGTH_HORIZONT+k] - zp,2))-ROBOT_RADIUS;
                    if (dist<RADIUS_A){
                        minDist=dist;
                        break;
                    }
                }
            }
        }
    }
    if(!(minDist < RADIUS_A)){
        if(so->nObstacles != 0){ //collision of robot with obstacles
            double type_o, d, x_o, y_o, obs_len, proj_len, dx, dy;
            for (int i = 0; i < so->nObstacles; i++) {
                x_o =double(so->obstacles[6*i+1]);
                y_o = double(so->obstacles[6*i+2]);
                d = double(so->obstacles[6*i+5]);
                type_o = double(so->obstacles[6*i]);
                for (int k = 0; k < LENGTH_HORIZONT; k++) {
                    if(type_o < 2){ //circle type
                        dist=sqrt(pow(so->states[k] - x_o,2) + pow(so->states[LENGTH_HORIZONT+k] - y_o,2)) - d;
                    }else{ //surface type
                        dx = double(so->obstacles[6*i+3]) - x_o;
                        dy = double(so->obstacles[6*i+4]) - y_o;
                        obs_len = (dx*dx)+(dy*dy);
                        proj_len = ((so->states[k]-x_o)*dx +(so->states[LENGTH_HORIZONT+k]-y_o)*dy)/(obs_len);
                        if(proj_len>1){
                            proj_len=1;
                        }else if(proj_len<0){
                            proj_len=0;
                        }
                        dx = x_o + proj_len*dx - so->states[k];//x_proj = x_o + proj_len*dx;
                        dy = y_o + proj_len*dy - so->states[LENGTH_HORIZONT+k];//y_proj = y_o + proj_len*dy;
                        dist = sqrt(dx*dx + dy*dy)-d;
                    }
                    if (dist<RADIUS_A){
                        minDist=dist;
                        break;
                    }
                }
            }
        }
    }
    *gj = RADIUS_A-minDist;
    return;
}

void LeaderSolver::callSolver (void) {
    int nparam,nf,nineq,neq,mode,iprint,miter,neqn,nineqn,
            ncsrl,ncsrn,nfsr,mesh_pts[1],inform;
    double bigbnd,eps,epsneq,udelta;
    double *bl,*bu,*f,*g,*lambda,*cd;
    mode=100;
    iprint=0; //printing informations
    miter=500; //maximum number of iterations
    bigbnd=1.e12; // all bounds above bigbnd or under -bigbnd in bu or bl are suposed to be infinite
    eps=1.e-9; //final norm requirement
    epsneq=1.e-6; // maximum allowed violation of nonlinear equality constraints
    udelta=0.e0; //perturbation size
    nparam=(N_PARAMS-2)*LENGTH_HORIZONT; //dimension of x
    nf=1; //number of objective functions
    neqn=neq=0; //3; //nonlinear equality constraints / equality constraints
    nineqn=1; //nonlinear inequality constraints / inequality constraints
    nineq=1;
    ncsrn=0; //sets of nonlinear sequentially related constraints
    ncsrl=0; //sets of linear sequentially related constraints
    mesh_pts[1]=0; //1
    nfsr=0; //sequentially related objective functions
    cd=(double *)calloc(nparam,sizeof(double));
    bl=(double *)calloc(nparam,sizeof(double)); //lower bounds
    bu=(double *)calloc(nparam,sizeof(double)); // upper bound
    f=(double *)calloc(nf+1,sizeof(double));
    g=(double *)calloc(nineq+(mesh_pts[1]-1)*(ncsrl+ncsrn)+neq+1,sizeof(double));
    lambda=(double *)calloc(nineq+(mesh_pts[1]-1)*(ncsrl+ncsrn)+neq+nf+nparam,sizeof(double));
    generateBoundaryValues(bl,bu);
    cfsqp(nparam,nf,nfsr,nineqn,nineq,neqn,neq,ncsrl,ncsrn,mesh_pts,
          mode,iprint,miter,&inform,bigbnd,eps,epsneq,udelta,bl,bu,(x),
          f,g,lambda,obj,cntr,grobfd,grcnfd,this);
    obj(5,5,x,&objf,this);
    free(bl);
    free(bu);
    free(f);
    free(g);
    free(lambda);
    free(cd);
}
