/**
    follower.cc
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#include <cmath>
#include <iostream>
#include <fstream>

#include "ros/ros.h"
#include "map.h"

using namespace std;
using namespace mkr::formation;

Map::Map(void) { }

Map::~Map(void) { }

void Map::loadMap(const char * name) {
  ifstream in(name);
  cout << " Loading MAP: " << endl;
  double x, y, z, hx, hy, hz, mode, t;//2 for surface, 1 for point
  Coordinates point;
  Coordinates pointOI;
  if (in.is_open()){
    while (in >> x >> y >> z >> hx >> hy >> hz >> mode) {
      point.x = x;
      point.y = y;
      point.z = z;
      pointOI.x = hx;
      pointOI.y = hy;
      pointOI.z = hz;
      t = atan2(hy-y, hx-x);
      pointOI.phi = t > 0 ? t : t+2*PI;
      pointsOfInterest.push_back(pointOI);
      coordinates.push_back(point);
      modes.push_back(mode);
    }
    cout << "	" << coordinates.size() << " segments of map have been loaded." << endl;
    cout << pointsOfInterest.size() << " segments of heading have been loaded" << endl;
    in.close();
  }else{
    cout << "Unable to open file";
    ros::shutdown();
  }
}

void Map::loadObstacles(const char * name) {
  ifstream in(name);
  double x, y, x2, y2, radius;
  int type;
  Obstacle obst;
  if (in.is_open()){
    while (in >> type >> x >> y >> x2 >> y2 >> radius) {
      obst.x = x;
      obst.y = y;
      obst.radius = radius;
      obst.type = type;
      obst.x2 = x2;
      obst.y2 = y2;
      obstacles.push_back(obst);
    }
    cout << "	" << obstacles.size() << " obstacles was loaded" << endl;
    in.close();
  } else{
    cout << "Unable to open file";
  }
}

VCoordinates & Map::getCoordinates(void) {
    return coordinates;
}

Obstacles & Map::getObstacles(void) {
    return obstacles;
}

VCoordinates & Map::getPointsOfInterest(void){
    return pointsOfInterest;
}
