/**
    robot.cc
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#include "robot.h"

using namespace std;
using namespace mkr::formation;

Time Robot::time;
Time Robot::runtime;

Robot::Robot(void) : input(0.0, 0.0, 0.0, 0.0, 0.0) {
      x = (double *)calloc(LENGTH_HORIZONT*N_PARAMS,sizeof(double));
      desiredTrajectory = (double *)calloc(LENGTH_HORIZONT*(N_PARAMS+1),sizeof(double));//+curvature
}

Robot::~Robot() {
    free(x);
    free(desiredTrajectory);
    if(obstacles != NULL){
       free(obstacles);
    }
}

void Robot::resetTime(void) {
	time.init();
}

void Robot::move(void){
    if (fabs(input.curvature) < (1E-6F)) {
    actualState[0]= actualState[0] + input.forwardVelocity*cos(actualState[2])*TS;
    actualState[1]= actualState[1] + input.forwardVelocity*sin(actualState[2])*TS;
    } else {
    actualState[0]= actualState[0] + (-sin(actualState[2])+sin(actualState[2]+input.curvature*input.forwardVelocity*TS)) / input.curvature;
    actualState[1]= actualState[1] - (-cos(actualState[2])+cos(actualState[2]+input.curvature*input.forwardVelocity*TS)) / input.curvature;
     }
     actualState[2] = actualState[2] + input.curvature*input.forwardVelocity*TS;
     actualState[3] = fmod(actualState[3] + input.angularVelocity*TS +2*PI, 2*PI);
     actualState[4] = input.forwardVelocity;
     actualState[5] = input.curvature;
     actualState[6] = input.angularVelocity;
     actualState[7] = actualState[7]+input.ascentVelocity*TS;
     actualState[8] = input.ascentVelocity;
     actualState[9] = actualState[9] + input.pitchVelocity*TS;
     actualState[10] = input.pitchVelocity;
}

void Robot::setObstacles(Obstacles obs){
    obstacles = (double*)calloc(6*obs.size(), sizeof(double));
    for(int i = 0; i < obs.size(); i++){
        obstacles[6*i] = obs[i].type;
        obstacles[6*i+1] = obs[i].x;
        obstacles[6*i+2] = obs[i].y;
        obstacles[6*i+3] = obs[i].x2;
        obstacles[6*i+4] = obs[i].y2;
        obstacles[6*i+5] = obs[i].radius;
    }
    nObstacle = obs.size();
}

double* Robot::getActualState(){
   return actualState;
}

